/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011, Robert Bosch LLC.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Robert Bosch nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************/

#ifndef JOINT_FORCE_CONTROLLER_H
#define JOINT_FORCE_CONTROLLER_H

/**
   @class pr2_controller_interface::JointForceController
   @brief Joint Force Controller (torque or force)

   This class provides a force controller which controls an arm joint effort towards a target value.


   @section ROS ROS interface

   @param type Must be "JointForceController"
   @param joint Name of the joint to control.

   Subscribes to:
   - @b command (std_msgs::Float64) : The joint effort to apply
*/

#include <boost/thread/condition.hpp>
#include <ros/node_handle.h>
#include <pr2_controller_interface/controller.h>
#include <control_toolbox/pid.h>
#include "control_toolbox/pid_gains_setter.h"
#include <boost/scoped_ptr.hpp>
#include <realtime_tools/realtime_publisher.h>

#include <std_msgs/Float64.h>
#include <pr2_controllers_msgs/JointControllerState.h>


namespace pr2_dremel
{

class JointForceController : public pr2_controller_interface::Controller
{
public:

  JointForceController();
  ~JointForceController();

  virtual bool init(pr2_mechanism_model::RobotState *robot, ros::NodeHandle &n);
  virtual void starting();
  virtual void update();
private:
  bool initialized_;

  pr2_mechanism_model::RobotState *robot_;       /**< Pointer to robot structure. */
  control_toolbox::Pid position_pid_controller_;          /**< Internal position PID controller. */
  control_toolbox::Pid velocity_pid_controller_;          /**< Internal velocity PID controller. */
  ros::Time last_time_;                          /**< Last time stamp of update. */

  boost::scoped_ptr<
    realtime_tools::RealtimePublisher<
      pr2_controllers_msgs::JointControllerState> > controller_state_publisher_ ;

  ros::NodeHandle node_;
  ros::Subscriber position_sub_;
  ros::Subscriber force_sub_;
  //ros::Publisher  vel_pub_;
 
  pr2_mechanism_model::JointState *joint_state_;
  bool   force_control_;
  double force_;
  double position_;
  double max_velocity_;
  double position_control_limit_;

  ros::Duration dt_;

  void forceCB(const std_msgs::Float64ConstPtr& msg);
  void positionCB(const std_msgs::Float64ConstPtr& msg);

  double position_effort(double position, ros::Time time);
  double velocity_effort(double velocity, ros::Time time);
  double force_effort(double force);

  double force_error(double force);
  double position_error(double position);

  static const int velocity_window_ = 30;
  int velocity_id_;
  float velocities_[velocity_window_]; 
  float velocity_sum_;  
  void reset_velocity_average();
};

}

#endif
