/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011, Robert Bosch LLC.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Robert Bosch nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************/
#include <ros/ros.h>
#include <pr2_dremel_server/Segment.h>
#include <geometry_msgs/Point.h>

using vlr::TagLine;
namespace dremel {

  class DrawObject{
  public:
    DrawObject(std::vector<std::vector<vlr::TagLine> > setlines, double initscale, double initx, double inity)
    {
      lines_=setlines;
      is_selected_=false;
      font_advance_=0.0;
      is_text_=true;
      updateCentroid();
    }
    
    DrawObject(std::vector<std::vector<vlr::TagLine> > setlines_)
    {
      lines_=setlines_;
      is_selected_=false;
      font_advance_=0.0;
      is_text_=true;
      updateCentroid();
    }

    ~DrawObject(){};
   
    void updateCentroid() {
      centroid_x_ = centroid_y_ = 0;
      int num_points = 0;
      std::vector<std::vector<vlr::TagLine> >& lines = getLines();
      for(std::vector<std::vector<vlr::TagLine> >::iterator itL=lines.begin();itL!=lines.end();++itL) {
        std::vector<vlr::TagLine>& tag_lines = *itL;
        for(std::vector<vlr::TagLine>::iterator itT=tag_lines.begin();itT!=tag_lines.end();++itT) {
          vlr::TagLine& tl = *itT;
          centroid_x_ += tl.x0();
          centroid_x_ += tl.x1();
          centroid_y_ += tl.y0();
          centroid_y_ += tl.y1();
          num_points+=2;
        }
      }
      centroid_x_ /= num_points;
      centroid_y_ /= num_points;
    }

    void move(float dx, float dy) {
      std::vector<std::vector<vlr::TagLine> >& lines = getLines();
      for(std::vector<std::vector<vlr::TagLine> >::iterator itL=lines.begin();itL!=lines.end();++itL) {
        std::vector<vlr::TagLine>& tag_lines = *itL;
        for(std::vector<vlr::TagLine>::iterator itT=tag_lines.begin();itT!=tag_lines.end();++itT) {
          vlr::TagLine& tl = *itT;
          tl.setX0(tl.x0()+dx);
          tl.setX1(tl.x1()+dx);
          tl.setY0(tl.y0()+dy);
          tl.setY1(tl.y1()+dy);
        }
      }
      updateCentroid();
    }

    void scale(float scale) {
      std::vector<std::vector<vlr::TagLine> >& lines = getLines();
      for(std::vector<std::vector<vlr::TagLine> >::iterator itL=lines.begin();itL!=lines.end();++itL) {
        std::vector<vlr::TagLine>& tag_lines = *itL;
        for(std::vector<vlr::TagLine>::iterator itT=tag_lines.begin();itT!=tag_lines.end();++itT) {
          vlr::TagLine& tl = *itT;
          tl.setX0((tl.x0()-centroid_x_)*scale + centroid_x_);
          tl.setX1((tl.x1()-centroid_x_)*scale + centroid_x_);
          tl.setY0((tl.y0()-centroid_y_)*scale + centroid_y_);
          tl.setY1((tl.y1()-centroid_y_)*scale + centroid_y_);
        }
      }
      updateCentroid();
    }

    bool isSelected(){
      return is_selected_;
    }

    void changeSelected(bool selected) {
      is_selected_ = selected;
    }

    void addToFontAdvance(float value){
      font_advance_+=value;
    }
    
    float getFontAdvance(){
      return font_advance_;
    }

    void setFontAdvance(float value){
      font_advance_=value;
    }

    void setAsText(){
      is_text_=true;
    }
    
    void setAsNotText(){
      is_text_=false;
    }

    bool isText(){
      return is_text_;
    }

    std::vector<std::vector<vlr::TagLine> >& getLines() {
     return lines_;
    }

    void insertNewLines(std::vector<std::vector<vlr::TagLine> > newlines){
      lines_.insert(lines_.end(), newlines.begin(), newlines.end());
      updateCentroid();
    }
    
    std::vector<pr2_dremel_server::Segment> convertToSegments(){
      std::vector<pr2_dremel_server::Segment> segments; 
             
       for(std::vector<std::vector<TagLine> >::iterator it=lines_.begin();it!=lines_.end();++it) {
         std::vector<TagLine>& line_segment = *it;
         std::vector<geometry_msgs::Point> points;
         pr2_dremel_server::Segment segment;

         for(std::vector<TagLine>::iterator itT=line_segment.begin();itT!=line_segment.end();++itT) {
           TagLine& line = *itT;
           geometry_msgs::Point p0,p1;
           p0.x = line.x0()*0.0005;
           p0.y = line.y0()*0.0005;
           p0.z=0;
           points.push_back(p0);
           p1.x = line.x1()*0.0005;
           p1.y = line.y1()*0.0005;
           p1.z=0;
           points.push_back(p1);
         }
         segment.points=points;
         segments.push_back(segment);
       }
       return segments;
    }
 
  private:
    std::vector<std::vector<vlr::TagLine> > lines_;
    bool is_selected_;
    float font_advance_;
    bool is_text_;
    float centroid_x_,centroid_y_;
  };

}

