/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011, Robert Bosch LLC.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Robert Bosch nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *********************************************************************/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

//== INCLUDES ==================================================================

// ros
#include <ros/ros.h>
#include <sensor_msgs/Image.h>
#include <sensor_msgs/image_encodings.h>
#include <sensor_msgs/CameraInfo.h>
#include <visualization_msgs/Marker.h>
#include <tf/tf.h>
#include <tf/transform_listener.h>
#include <pr2_dremel_server/CarveSegmentsAction.h>
#include <actionlib/client/simple_action_client.h>


// qt
#include <QMainWindow>
#include <QToolBox>

// display lib
#include <vlr/displayGL.h>

// local
#include "menu_widget.h"
#include "draw_object.h"

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_STROKER_H


//== CLASS DEFINITION ==========================================================
class MainWindow : public QMainWindow
{

    // Call Qt Meta Object Compiler
    Q_OBJECT

public:

    //-----------------------------------------------------------------------------
    //! Constructors
    MainWindow();

    //-----------------------------------------------------------------------------
    //! Destructor
    ~MainWindow();

    //! convert font glyphs to tag lines
    float glyphToLines(wchar_t ch, int size, float outline_width, std::vector<std::vector<vlr::TagLine> >& all_lines, float initial_font_advance);

    //! convert tag lines to gcode
    void linesToGcode(std::vector<std::vector<vlr::TagLine> >& lines);

    void sendLineMarker(std::vector<std::vector<vlr::TagLine> >& lines, std::string frame_id, double r, double g, double b);
    void addCountoursToImage(std::vector<std::vector<vlr::TagLine> >& contours);
    std::pair<dremel::DrawObject*,float> getClosestObject(float x, float y);
    void updateImage();
    // void updateImageLine();
    void moveDrawObject(dremel::DrawObject* object, int dx, int dy);
    void scaleDrawObject(dremel::DrawObject* object, float scale);
    void eraseObject(dremel::DrawObject* object);

public slots:
    void mousePress(int x, int y, uint8_t buttons);
    void mouseRelease(int x, int y, uint8_t buttons);
    void mouseMove(int x, int y, uint8_t buttons);
    void keyPress(unsigned int key);
    void mouseWheel(int x, int y, uint8_t buttons, int delta);
    //! Send segments for carving
    bool send();
    //! Cancel carving
    void cancel();

    void enableDremel();
    void disableDremel();
    void enableVacuum();
    void disableVacuum();

private:
    int32_t width_, height_;

    //! ros stuff
    ros::NodeHandle nh_;
    tf::TransformListener tf_listener_;
    ros::Rate loop_rate_;
    ros::Subscriber plate_img_sub_;
    sensor_msgs::Image plate_img_;

    //! widgets
    vlr::DisplayGL* display_widget_;
    //! Toolbox widget
    QToolBox *right_toolbox_;
    //! Menu Widget
    MenuWidget *menu_widget_;

    std::vector<dremel::DrawObject> draw_objects_;
    dremel::DrawObject* current_object_;


    float current_carving_progress_;
    vlr::Image<uint8_t>* img_;
    vlr::TagPoint* current_pt_;
    static const int32_t catch_dist_;
    std::vector<std::vector<vlr::TagLine> > lines_;
    int last_move_x_,last_move_y_;
    double table_height_;
    int last_right_x_,last_right_y_;

    // font stuff
    FT_Library library_;
    FT_Face face_;
    std::string fontfile_name_;
    uint8_t* font_buffer_;

    actionlib::SimpleActionClient<pr2_dremel_server::CarveSegmentsAction> action_client_;

    // visualization markers
    visualization_msgs::Marker lines_marker_;
    ros::Publisher viz_marker_pub_;
    int current_marker_id_;

    void feedbackCb(const pr2_dremel_server::CarveSegmentsFeedbackConstPtr& feedback){
      current_carving_progress_=feedback->percent_complete;
      menu_widget_-> carving_progress_->setValue(current_carving_progress_*100);
      printf("Percent Complete %f\n", current_carving_progress_);
    }

};

//==============================================================================
#endif // MAINWINDOW_H defined
//==============================================================================
